# read tob1 and ghg files.
from pathlib import Path
import numpy as np
import pandas as pd
import zipfile
import re

class Flux_reader(object):
    """
    acknowledgement: USDA-ARS-USSL 
    citation: 
        T. H. Skaggs, R. G. Anderson, J. G. Alfieri, T. M. Scanlon, W. P. Kustas (2018). 
        Fluxpart: Open source software for partitioning carbon dioxide and water vapor fluxes. 
        Agricultural and Forest Meteorology 253--254:218--224, 
        doi:10.1016/j.agrformet.2018.02.019.
    url: https://github.com/usda-ars-ussl/fluxpart/blob/master/fluxpart/util.py
    """
    def __init__(self, data_folder, MUTE = True):
        super().__init__()
        # initialize params: {
        self.MUTE = MUTE
        self.TYPE_MAP = {
            "IEEE4": np.float32,
            "IEEE8": np.float64,
            "LONG": np.int32,
            "ULONG": np.uint32,
        }
        # }

        # dir of data:
        self.data_folder = Path(data_folder)
    
    def __call__(self, data_type = "tob1"):
        if data_type == "tob1":
            if not self.MUTE: print("processing tob1 files...")
            paths = self.data_folder.glob(r"*.dat")
            dfs_generator = self.read_tob1(paths)
        elif data_type == "ghg":
            if not self.MUTE: print("processing ghg files...")
            paths = self.data_folder.glob(r"*.ghg")
            dfs_generator = self.read_ghg(paths)
        else:
            raise Exception("Wrong flux data type.")
        return dfs_generator

    def read_tob1(self, tobfiles, count = -1):
        """read tob1 files to dataframes."""
        for tobfile in tobfiles:
            with open(tobfile, "rb") as f:
                f.readline()
                names = f.readline().decode().strip().replace('"', "").split(",")
                f.readline()
                f.readline()
                types = f.readline().decode().strip().replace('"', "").split(",")
                dtype = np.dtype([(n, self.TYPE_MAP[t]) for n, t in zip(names, types)])
                array = np.fromfile(f, dtype=dtype, count=count)
                df = pd.DataFrame(array)
                yield df

    def read_ghg(self, ghg_files, *args, **kwargs):
        """Buffered pd.read_csv of data split across multiple files."""
        for ghg_file in ghg_files:
            with zipfile.ZipFile(ghg_file) as z:
                meta_basename = ghg_file.with_suffix(".metadata").name
                try:
                    with z.open(meta_basename) as f:
                        lines = f.readlines()
                        # convert string to dict:{
                        lines = [line.decode().strip().split("=") for line in lines if b"=" in line]
                        # use results to make a dict
                        lines = dict(lines)

                        skip_rows = int(lines["header_rows"]) - 1
                        sep = lines["separator"]
                        if sep == "tab":
                            sep = "\t"
                        else:
                            sep = ","
                        # }
                except Exception as e:
                    if not self.MUTE: print(e)
                basename = ghg_file.with_suffix(".data").name
                try:
                    df = pd.read_csv(z.open(basename), sep = sep, skiprows = skip_rows, *args, **kwargs)
                except Exception as e:
                    if not self.MUTE: print(e.args[0])
                    continue
            if isinstance(df, pd.io.parsers.TextFileReader):
                yield from df
            else:
                yield df   


# """
# example:
# """
# def main():
#     path = "data_ghg"
#     flux_read = Flux_reader(path, MUTE = False)
#     dfs = flux_read(data_type = "ghg")
#     for i, df in enumerate(dfs):
#         # print(df)
#         df.to_csv("ghg_" + str(i).zfill(3) + ".csv")

# if __name__ == "__main__":
#     main()



class EddyProc(object):
    super(object)
    def __init__(self):
        # Constants and default parameters.
        # self.offset1 = 0
        # self.gain1 = 1
        # self.curvature1 = 0
        pass

    # Choices for processing list
    def location_output_files(self, dCalc = "", dSpec = "", dWave = "", dCroCor = "", 
                            dDis = "", dQuad = "", dRef = ""):
        return (
            f"Location Output Files\n\t"
                f"Output File Calculations = {dCalc}\n\t"
                f"Output File Spectral = {dSpec}\n\t"
                f"Output File Wavelet = {dWave}\n\t"
                f"Output File Cross Correlation = {dCroCor}\n\t"
                f"Output File Distribution = {dDis}\n\t"
                f"Output File Quadrant = {dQuad}\n\t"
                f"Output File Reference = {dRef}\n"
        )

    def comments(self, cmt1 = "", cmt2 = "", cmt3 = ""):
        return (
            f"Comments\n\t"
                f"Comment = {cmt1}\n\t"
                f"Comment = {cmt2}\n\t"
                f"Comment = {cmt3}\n"
        )

    def extract(self, sTime = "", eTime = "", chn = None, label = ""):
        return (
            f"Extract\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Channel = {chn}\n\t"
                f"Label for Signal = {label}\n"
        )

    def gas_conversion_time_series(self, sTime = "", eTime = "", sig = "", 
                                    conFrom = "", conTo = "",
                                    offset1 = "0", gain1 = "1", curvature1 = "0", sigTC = "",
                                    valTC = "", sigPkPa = "", valPkPa = "", sigH2O = "", 
                                    valH2O = "", unitsH2O = "", moleWeight = "18",
                                    offset2 = "0", gain2 = "1", curvature2 = "0"):
        return(
            f"Gas conversion time series\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal = {sig}\n\t"
                f"Convert from = {conFrom}\n\t"
                f"Convert to = {conTo}\n\t"
                f"1st Offset = {offset1}\n\t"
                f"1st Gain = {gain1}\n\t"
                f"1st Curvature = {curvature1}\n\t"
                f"Signal T, C = {sigTC}\n\t"
                f"Value T, C = {valTC}\n\t"
                f"Signal P, kPa = {sigPkPa}\n\t"
                f"Value P, kPa = {valPkPa}\n\t"
                f"Signal H2O = {sigH2O}\n\t"
                f"Value H2O = {valH2O}\n\t"
                f"Units H2O = {unitsH2O}\n\t"
                f"Molecular Weight = {moleWeight}\n\t"
                f"2nd Offset = {offset2}\n\t"
                f"2nd Gain = {gain2}\n\t"
                f"2nd Curvature = {curvature2}\n"     
        )

    def despike(self, sTime = "", eTime = "", sig = "", std = "5", spikeWid = "4", consistency = "30",
                replaceSpikes = "", SLScount = "", outlierStd = "9"):
        return(
            f"Despike\n\t"
                f"From Time = {sTime}\n\t" 
                f"To Time = {eTime}\n\t" 
                f"Signal = {sig}\n\t"
                f"Standard Deviations = {std}\n\t"
                f"Spike width = {spikeWid}\n\t"
                f"Spike % consistency = {consistency}\n\t"
                f"Replace spikes = {replaceSpikes}\n\t"
                f"Storage Label spike count = {SLScount}\n\t"
                f"Outlier Standard Deviations = {outlierStd}\n"      
        )

    def despike_vickers(self, sTime = "", eTime = "", sig = "", winSize = "5", winStep = "10", 
                        iniStd = "3.5", increStd = "0.2", maxSpikeWid = "3", maxNum = "10",
                        count_ts_vickers = "", SLDSpikes = ""):
        return(
            f"Despike - Vickers\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {sTime}\n\t" 
                f"Signal = {sig}\n\t"
                f"Window size, minutes = {winSize}\n\t"
                f"Window step, points =  {winStep}\n\t"
                f"Initial standard deviation = {iniStd}\n\t"
                f"Standard deviation increment, % = {increStd}\n\t"
                f"Max spike width = {maxSpikeWid}\n\t"
                f"Max number of passes = {maxNum}\n\t"
                f"Storage Label spike count = {count_ts_vickers}\n\t"
                f"Storage Label data % spikes = {SLDSpikes}\n"
        )

    def one_chn_statistics(self, sTime = "", eTime = "", sig = "", SLmean = "", SLStd = "", 
                            SLSkew = "", SLKurt = "", SLMax = "", SLMin = "", 
                            SLVar = "", SLTurInten = "", ATIDenomi = ""):
        return(
            f"1 chn statistics\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal = {sig}\n\t"
                f"Storage Label Mean = {SLmean}\n\t"
                f"Storage Label Std Dev = {SLStd}\n\t"
                f"Storage Label Skewness = {SLSkew}\n\t"
                f"Storage Label Kurtosis = {SLKurt}\n\t"
                f"Storage Label Maximum = {SLMax}\n\t"
                f"Storage Label Minimum = {SLMin}\n\t" 
                f"Storage Label Variance = {SLVar}\n\t"
                f"Storage Label Turbulent Intensity = {SLTurInten}\n\t"
                f"Alt Turbulent Intensity Denominator = {ATIDenomi}\n"        
        )

    def virtual_temperature_raw(self, sTime = "", eTime = "", sig = "", sigH2O = "", PKpa = "",
                                wVapourUnit = "Absolute density g/m3", tempConver = "Calculate true from virtual-sonic"):
        return(
            f"Virtual Temperature Raw\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal T(C) = {sig}\n\t"
                f"Signal H2O = {sigH2O}\n\t"
                f"Pressure, kPa = {PKpa}\n\t"
                f"Water vapour units = {wVapourUnit}\n\t"
                f"Temperature conversion = {tempConver}\n"        
        )

    def plot_value(self, sTime = "", eTime = "", LVal = "", RVal = "", LMin = "", LMax = "",
                    RMin = "", RMax = "", match = ""):
        return(
            f"Plot Value\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Left Axis Value = {LVal}\n\t"
                f"Right Axis Value = {RVal}\n\t"
                f"Left Axis Minimum = {LMin}\n\t"
                f"Left Axis Maximum = {LMax}\n\t"
                f"Right Axis Minimum = {RMin}\n\t"
                f"Right Axis Maximum = {RMax}\n\t"
                f"Match Left/Right Axes = {match}\n"
        )

    def wind_direction(self, sTime = "", eTime = "", sigU = "", sigV = "", orient = "",
                        windDirCompo = "U+N_V+W", windDirOut = "N_0_deg-E_90_deg",
                        SLWindDir = "", SLWindDirDiv = ""):
        return(
            f"Wind direction\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal (u) = {sigU}\n\t"
                f"Signal (v) = {sigV}\n\t"
                f"Orientation = {orient}\n\t"
                f"Wind Direction Components = {windDirCompo}\n\t"
                f"Wind Direction Output = {windDirOut}\n\t"
                f"Storage Label Wind Direction = {SLWindDir}\n\t"
                f"Storage Label Wind Dir Std Dev = {SLWindDirDiv}\n"
        )

    def cross_correlate(self, sTime = "", eTime = "", sig = "", sigLag = "", corrType = "Covariance", 
                        corrCurveOut = "", SLPeakTime = "", SLPeakVal = ""):
        return(
            f"Cross Correlate\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal = {sig}\n\t"
                f"Signal which lags = {sigLag}\n\t"
                f"Correlation type = {corrType}\n\t"
                f"Output Correlation curve = {corrCurveOut}\n\t"
                f"Storage Label Peak Time = {SLPeakTime}\n\t"
                f"Storage Label Peak Value = {SLPeakVal}\n"
        )

    def plot_cross_correlation(self, sTime = "", eTime = ""):
        return(
            f"Plot cross correlation\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n"
        )

    def remove_lag(self, sTime = "", eTime = "", sig = "", minLagS = "", lagS = "", maxLagS = "",
                    belowMinDefault = "", belowMaxDefault = ""):
        return(
            f"Remove Lag\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal = {sig}\n\t"
                f"Min Lag (sec) =  {minLagS}\n\t"
                f"Lag (sec) =  {lagS}\n\t"
                f"Max Lag (sec) = {maxLagS}\n\t"
                f"Below Min default (sec) = {belowMinDefault}\n\t"
                f"Above Max default (sec) = {belowMaxDefault}\n"
        )

    def rotation_coefficients(self, sTime = "", eTime = "", sigU = "", sigV = "", sigW = "", 
                                SLAlpha = "alpha", SLBeta = "beta", SLGamma = "gamma",
                                optMeanU = "", optMeanV = "", optMeanW = ""):
        return(
            f"Rotation coefficients\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal (u) = {sigU}\n\t"
                f"Signal (v) = {sigV}\n\t"
                f"Signal (w) = {sigW}\n\t"
                f"Storage Label Alpha = {SLAlpha}\n\t"
                f"Storage Label Beta  = {SLBeta}\n\t"
                f"Storage Label Gamma = {SLGamma}\n\t"
                f"Optional mean u = {optMeanU}\n\t"
                f"Optional mean v = {optMeanV}\n\t"
                f"Optional mean w = {optMeanW}\n"
        )

    def rotation(self, sTime = "", eTime = "", sigU = "", sigV = "", sigW = "",
                alpha = "alpha", beta = "beta", gamma = "gamma",
                do1Rot = "", do2Rot = "", do3Rot = ""):
        return(
            f"Rotation\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal (u) = {sigU}\n\t"
                f"Signal (v) = {sigV}\n\t"
                f"Signal (w) = {sigW}\n\t"
                f"Alpha = {alpha}\n\t"
                f"Beta = {beta}\n\t"
                f"Gamma = {gamma}\n\t"
                f"Do 1st Rot = {do1Rot}\n\t"
                f"Do 2nd Rot = {do2Rot}\n\t"
                f"Do 3rd Rot = {do3Rot}\n"
        )

    def spectra(self, sTime = "", eTime = "", sig = "", wind = "Hamming", kaiser_bell = "", 
                specOut = "x", sortOut = "x", detrendData = "Low"):
        return(
            f"Spectra\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal = {sig}\n\t"
                f"Window = {wind}\n\t"
                f"Kaiser/Bell coef = {kaiser_bell}\n\t"
                f"Output Spectra = {specOut}\n\t"
                f"Sort Output = {sortOut}\n\t"
                f"Detrend data = {detrendData}\n"
        )

    def cospectra(self, sTime = "", eTime = "", sig1 = "", sig2 = "", wind = "Hamming", kaiser_bell = "", 
                cospecOut = "x", cohereOut = "", phaseOut = "", sortOut = "x", 
                detrendSig1 = "Low", detrendSig2 = "Low"):
        return(
            f"Cospectra\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Signal 1 = {sig1}\n\t"
                f"Signal 2 = {sig2}\n\t"
                f"Window = {wind}\n\t"
                f"Kaiser/Bell = {kaiser_bell}\n\t"
                f"Output Cospectra = {cospecOut}\n\t"
                f"Output Coherence = {cohereOut}\n\t"
                f"Output Phase = {phaseOut}\n\t"
                f"Sort Output = {sortOut}\n\t"
                f"Detrend data, Signal 1 = {detrendSig1}\n\t"
                f"Detrend data, Signal 2 = {detrendSig2}\n" 
        )

    def plot_spectral(self, sTime = "", eTime = "", LAxSpec = "Spectra", RAxCospec = "Cospectra", 
                    LAxLog = "x", RAxLog = "x", LAxMin = "", LAxMax = "", RAxMin = "", RAxMax = "", matchAxs = ""):
        return(
            f"Plot spectral\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Left Axis Spectra = {LAxSpec}\n\t"
                f"Right Axis Spectra = {RAxCospec}\n\t"
                f"Left Axis Logarithmic = {LAxLog}\n\t"
                f"Right Axis Logarithmic = {RAxLog}\n\t"
                f"Left Axis Minimum = {LAxMin}\n\t"
                f"Left Axis Maximum = {LAxMax}\n\t"
                f"Right Axis Minimum = {RAxMin}\n\t"
                f"Right Axis Maximum = {RAxMax}\n\t"
                f"Match Left/Right Axes = {matchAxs}\n"   
        )

    def sensible_heat_flux_coefficient(self, sTime = "", eTime = "", SLable = "", appTo = "", appBy = "",
                                        vaporPres = "", minQC = "", maxQC = "", temp = "",
                                        pres = "", altrhoCp = ""):
        return(
            f"Sensible heat flux coefficient\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {sTime}\n\t"
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t"
                f"Apply by = {appBy}\n\t"
                f"Vapour pressure (KPa) = {vaporPres}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"Temperature (C) = {temp}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"Pressure (KPa) = {pres}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"Alternate rhoCp = {altrhoCp}\n" 
        )

    def latent_heat_of_evaporation(self, sTime = "", eTime = "", SLable = "", appTo = "", appBy = "",
                                    temp = "", minQC = "", maxQC = "", pres = "", LECoef = ""):
        return(
            f"Latent heat of evaporation\n\t"
                f"From Time = {sTime}\n\t" 
                f"To Time = {eTime}\n\t" 
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t" 
                f"Apply by = {appBy}\n\t" 
                f"Temperature (C) = {temp}\n\t"
                f"Min or QC = {minQC}\n\t" 
                f"Max or QC = {maxQC}\n\t" 
                f"Pressure (KPa) = {pres}\n\t"
                f"Min or QC = {minQC}\n\t" 
                f"Max or QC = {maxQC}\n\t" 
                f"LE flux coef, L = {LECoef}\n" 
        )

    def friction_velocity(self, sTime = "", eTime = "", sigU = "", sigV = "", sigW = "",
                        SLuw = "", SLuwvw = ""):
        return(
            f"Friction Velocity\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t" 
                f"Signal (u) = {sigU}\n\t"
                f"Signal (v) = {sigV}\n\t"
                f"Signal (w) = {sigW}\n\t"
                f"Storage Label U* (uw) = {SLuw}\n\t"
                f"Storage Label U* (uw vw) = {SLuwvw}\n"
        )

    def two_chn_statistics(self, sTime = "", eTime = "", sig1 = "", sig2 = "", SLCov = "", SLCor = "",
                            SLFlux = "", fluxCoef = ""):
        return(
            f"2 chn statistics\n\t"
                f"From Time = {sTime}\n\t" 
                f"To Time = {eTime}\n\t" 
                f"Signal = {sig1}\n\t"
                f"Signal = {sig2}\n\t"
                f"Storage Label Covariance = {SLCov}\n\t"
                f"Storage Label Correlation = {SLCor}\n\t" 
                f"Storage Label Flux = {SLFlux}\n\t"
                f"Flux coefficient = {fluxCoef}\n"
        )

    def tube_attenuation(self, sTime = "", eTime = "", SLable = "", appTo = "", appBy = "",
                        gasSpecies = "", tubePres = "", minQC = "", maxQC = "", flowRate = "",
                        tubeLen = "", tubeID = "", uDefLambdaCoef = ""):
        return(
            f"Tube attenuation\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t"
                f"Apply by = {appBy}\n\t"
                f"Gas species = {gasSpecies}\n\t"
                f"Tube pressure (KPa) = {tubePres}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"Flow rate (LPM) = {flowRate}\n\t"
                f"Tube length (m) = {tubeLen}\n\t"
                f"Tube ID (m) = {tubeID}\n\t"
                f"User defined Lambda coefficient = {uDefLambdaCoef}\n"
        )

    def set_value(self, valDict = {}, sTime = "", eTime = ""):
        items = valDict.items()
        varNum = len(items)
        retLines = (
            f"Set Values\n\t"
                f"From Time = {sTime}\n\t" 
                f"To Time = {sTime}\n\t"
                f"Number of Variables = {varNum}\n\t"
        )
        # print(type(retLines))
        for item in items:
            retLines += (
                f"Storage Label = {item[0]}\n\t" 
                f"Assignment value = {item[1]}\n\t"
            )
        ## \t should be cut out from the last line.
        # retLines = retLines[0: -1]
        return retLines[0: -1]

    def mathematical_operation(self, sTime = "", eTime = "",  SLable = "", appTo = "", appBy = "",
                                measuredVarA = "", operation = "", measuredVarB = ""):
        return(
            f"Mathematical operation\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t"
                f"Apply by = {appBy}\n\t"
                f"Measured variable A =  {measuredVarA}\n\t"
                f"Operation  = {operation}\n\t"
                f"Measured variable B =  {measuredVarB}\n"
        )

    def user_defined(self, sTime = "", eTime = "",  SLable = "", appTo = "", appBy = "",
                    equation = "", var = ""):
        return(
            f"User defined\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t" 
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t"
                f"Apply by = {appBy}\n\t" 
                f"Equation = {equation}\n\t"
                f"Variable = {var}\n\t"
        )

    def stability_monin_obhukov(self, sTime = "", eTime = "",  SLable = "", appTo = "", appBy = "",
                                measHeight = "", zeroPlaneDisp = "", virtualTemp = "",
                                minQC = "", maxQC = "", HFlux = "", HFluxCoef = "", fVelo = ""):
        return(
            f"Stability - Monin Obhukov\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t"
                f"Apply by = {appBy}\n\t" 
                f"Measurement height (m) =  {measHeight}\n\t"
                f"Zero plane displacement (m) = {zeroPlaneDisp}\n\t"
                f"Virtual Temperature (C) = {virtualTemp}\n\t"
                f"Min or QC = {minQC}\n\t" 
                f"Max or QC = {maxQC}\n\t"
                f"H flux (W/m2) = {HFlux}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"H flux coef, RhoCp = {HFluxCoef}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC = {maxQC}\n\t"
                f"Scaling velocity (m/s) =  {fVelo}\n\t"
                f"Min or QC = {minQC}\n\t"
                f"Max or QC =  {maxQC}\n"       
        )

    def frequency_response(self, sTime = "", eTime = "",  SLable = "", appTo = "", appBy = "",
                        corrType = "", measHeight = "", zeroPlaneDisp = "", boundLyrHeight = "",
                        stability = "", windSpeed = "", 
                        sen1FlowVelo = "", sen1SampFreq = "",
                        sen1LPassFiltType = "", sen1LPassFiltTimeConstant = "",
                        sen1HPassFiltType = "", sen1HPassFiltTimeConstant = "",
                        sen1PathLen = "", sen1TimeConstant = "", sen1TubeAttenuCoef = "",
                        sen2FlowVelo = "", sen2SampFreq = "",
                        sen2LPassFiltType = "", sen2LPassFiltTimeConstant = "",
                        sen2HPassFiltType = "", sen2HPassFiltTimeConstant = "",
                        sen2PathLen = "", sen2TimeConstant = "", sen2TubeAttenuCoef = "",
                        pathSepera = "", getSpecDataType = "", getRespFuncFrom = "", 
                        refTag = "", refRespCond = "", sen1SubSamp = "", sen2SubSamp = "",
                        appVeloDistAdjust = "", useCalcDist = "", veloDistStd = "", stabilityDistStd = ""):
        return(
            f"Frequency response\n\t"
                f"From Time = {sTime}\n\t"
                f"To Time = {eTime}\n\t"
                f"Storage Label = {SLable}\n\t"
                f"Apply to = {appTo}\n\t" 
                f"Apply by = {appBy}\n\t"
                f"Correction type = {corrType}\n\t"
                f"Measurement height (m) = {measHeight}\n\t"
                f"Zero plane displacement (m) = {zeroPlaneDisp}\n\t"
                f"Boundary layer height (m) = {boundLyrHeight}\n\t"
                f"Stability Z/L = {stability}\n\t"
                f"Wind speed (m/s) = {windSpeed}\n\t"
                f"Sensor 1 Flow velocity (m/s) = {sen1FlowVelo}\n\t"
                f"Sensor 1 Sampling frequency (Hz) = {sen1SampFreq}\n\t"
                f"Sensor 1 Low pass filter type = {sen1LPassFiltType}\n\t"
                f"Sensor 1 Low pass filter time constant = {sen1LPassFiltTimeConstant}\n\t" 
                f"Sensor 1 High pass filter type = {sen1HPassFiltType}\n\t"
                f"Sensor 1 High pass filter time constant = {sen1HPassFiltTimeConstant}\n\t"
                f"Sensor 1 Path length (m) = {sen1PathLen}\n\t"
                f"Sensor 1 Time constant (s) = {sen1TimeConstant}\n\t"
                f"Sensor 1 Tube attenuation coef = {sen1TubeAttenuCoef}\n\t" 
                f"Sensor 2 Flow velocity (m/s) = {sen2FlowVelo}\n\t"
                f"Sensor 2 Sampling frequency (Hz) = {sen2SampFreq}\n\t"
                f"Sensor 2 Low pass filter type = {sen2LPassFiltType}\n\t"
                f"Sensor 2 Low pass filter time constant = {sen2LPassFiltTimeConstant}\n\t"
                f"Sensor 2 High pass filter type = {sen2HPassFiltType}\n\t"
                f"Sensor 2 High pass filter time constant = {sen2HPassFiltTimeConstant}\n\t"
                f"Sensor 2 Path length (m) = {sen2PathLen}\n\t"
                f"Sensor 2 Time constant (s) = {sen2TimeConstant}\n\t"
                f"Sensor 2 Tube attenuation coef = {sen2TubeAttenuCoef}\n\t"
                f"Path separation (m) = {pathSepera}\n\t"
                f"Get spectral data type = {getSpecDataType}\n\t"
                f"Get response function from = {getRespFuncFrom}\n\t"
                f"Reference Tag = {refTag}\n\t"
                f"Reference response condition = {refRespCond}\n\t" 
                f"Sensor 1 subsampled = {sen1SubSamp}\n\t"
                f"Sensor 2 subsampled = {sen2SubSamp}\n\t"
                f"Apply velocity distribution adjustment = {appVeloDistAdjust}\n\t" 
                f"Use calculated distribution = {useCalcDist}\n\t" 
                f"Velocity distribution std dev= {veloDistStd}\n\t" 
                f"Stability distribution std dev= {stabilityDistStd}\n" 
        )


import os
import ee
import folium
import requests
import numpy as np
from pathlib import Path

class EEarth(object):
	"""
	Earth engine object
	"""
	def __init__(self, dataset_name):
		super()
		ee.Initialize()
		self.dataset_name = dataset_name

	def fetch_image(self):
		self.image = ee.Image(self.dataset_name)

	def fetch_collection(self, date_range = [], roi = None, to_list = False):
		collection = ee.ImageCollection(self.dataset_name)
		# filter by date:
		if date_range:
			start_date, end_date = date_range
			collection = collection.filterDate(start_date, end_date)
		# filter by bounds:
		if roi:
			collection = collection.filterBounds(roi)
		self.__length__ = collection.size().getInfo()
		# covert ee collection to list?
		if to_list: 
			collection = collection.toList(collection.size())
		self.collection = collection
	
	def get_image_by_index(self, idx):
		self.image = ee.Image(self.collection.get(idx))

	def reduce_collection(self, band = None, label = "DATA", band_reducer = ee.Reducer.mean(), spatial_reducer = ee.Reducer.median()):
		if band:
			image = self.collection.select(band).reduce(spatial_reducer).rename(band)
		else:
			image = self.collection.map(
					lambda image: image.reduce(band_reducer)
				).reduce(spatial_reducer).rename(label)
		self.image = image

class Emagebox(object):
	"""
	Emagebox: Earth engine Image Toolbox
	"""
	def __init__(self, image):
		self.image = image

	def get_band_names(self):
		return self.image.bandNames().getInfo()

	def get_stats(self, roi, reducer = ee.Reducer.mean(), scale = 30, max_pixels = 1e8):
		stat = self.image.reduceRegion(
			reducer = reducer,
			geometry = roi,
			scale = scale,
			maxPixels = max_pixels
		)
		return stat.getInfo()
	
	def get_date(self):
		date = ee.Date(self.image.get('system:time_start'))
		return date.format('Y-M-d').getInfo()
	
	def get_proj(self, band, get_info = False):
		# // Get projection information from band.
		proj = self.image.select(band).projection()
		if get_info:
			proj = proj.getInfo()
		return proj

	def get_scale(self, band):
		# // Get scale (in meters) information from band.
		scale = self.image.select(band).projection().nominalScale()
		return scale.getInfo()

	def reproject(self, proj):
		return self.image.reproject(proj)

	def clip(self, roi):
		return self.image.clip(roi)

	def mask_value(self, value = -9999):
		mask = self.image.eq(value)
		self.image = self.image.updateMask(mask)

	def unmask(self, value = -9999):
		self.image = self.image.unmask(ee.Image.constant(value))

	def get_value(self, band, point, scale = 10, reducer = ee.Reducer.first()):
		value = self.image.select(band).reduceRegion(reducer, point, scale).get(band)
		value = ee.Number(value)
		return value.getInfo()

	def get_values(self, band, roi, unmask = -9999, max_pixels = 1e8, scale = 20):
		# getInfo() is limited to 5000 records
		# ee.ee_exception.EEException: Array: No numbers in 'values', must provide a type.
		# see more at: https://gis.stackexchange.com/questions/321560/getting-dem-values-as-numpy-array-in-earth-engine
		# for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
		# for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
		latlng = ee.Image.pixelLonLat().addBands(self.image.clip(roi).unmask(ee.Image.constant(unmask)))
		latlng = latlng.reduceRegion(
			reducer = ee.Reducer.toList(), 
			geometry = roi, 
			maxPixels = max_pixels, 
			scale = scale
		)
		lats = np.array((ee.Array(latlng.get("latitude")).getInfo()))
		lons = np.array((ee.Array(latlng.get("longitude")).getInfo()))
		try:
			values = np.array((ee.Array(latlng.get(band)).getInfo()))
		except:
			values = np.full_like(lats, np.nan, dtype = np.float64)
		# self.values = list(values) ## print as list to check

		if not (values.shape == lats.shape == lons.shape):
			raise Exception(
				f"SizeError: " +
				f"values shape is {values.shape}, " + 
				f"lats shape is {lats.shape}, " + 
				f"lons shape is {lons.shape}."
			)
		return {
			"values":  values,
			"lons": lons,
			"lats": lats
		}

	def localize(self, save_name, save_folder = ".", crs_epsg = "4326", scale = 1000):
		url = self.image.getDownloadURL({
			"name": save_name,
			"crs": "EPSG:" + crs_epsg,
			"scale": scale
		})
		if not os.path.exists(save_folder): os.makedirs(save_folder)
		save_dir = f"{save_folder}/{save_name}.zip"

		# Download the subset
		r = requests.get(url, stream = True)
		with open(save_dir, 'wb') as fd:
			for chunk in r.iter_content(chunk_size = 1024):
				fd.write(chunk)

class Utils(object):
	def __init__(self):
		pass

	@classmethod
	def coors2roi(self, bounds):
		# for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
		# for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
		if isinstance(bounds, list):
			if np.array(bounds).ndim == 1:
				roi = ee.Geometry.Rectangle(bounds)
			else:
				roi = ee.Geometry.Polygon(bounds)
		return roi

	@classmethod
	def coor2point(self, coors = None, lon = None, lat = None):
		if coors:
			lon, lat = coors
		assert lon, "No longitude input..."
		assert lat, "No latitude input..."
		point = ee.Geometry.Point(lon, lat)
		return point

	@classmethod
	def get_circle_buffer(self, point, buffer_size = 100):
		return point.buffer(buffer_size)

	@classmethod	
	def get_rect_buffer(self, point, buffer_size = 0.5):
		# buffer_size unit eq proj unit, e.g., degree for WGS84
		lon, lat = point.getInfo()["coordinates"]
		# example: [-97.94, 26.81, -96.52, 26.84]
		bounds = [lon - buffer_size, lat - buffer_size, lon + buffer_size, lat + buffer_size]
		return ee.Geometry.Rectangle(bounds)
	

class Canvas(object):
	def __init__(self):
		from folium import plugins
		self.cmap = {
			"veg": ['FFFFFF', 'CE7E45', 'DF923D', 'F1B555', 'FCD163', '99B718', '74A901',
					'66A000', '529400', '3E8601', '207401', '056201', '004C00', '023B01',
					'012E01', '011D01', '011301'
					],
			"grey": ['006633', 'E5FFCC', '662A00', 'D8D8D8', 'F5F5F5']
		}
		# Add custom basemaps to folium
		self.basemaps = {
			'Google Maps': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Maps',
				overlay = True,
				control = True
			),
			'Google Satellite': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Satellite',
				overlay = True,
				control = True
			),
			'Google Terrain': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Terrain',
				overlay = True,
				control = True
			),
			'Google Satellite Hybrid': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Satellite',
				overlay = True,
				control = True
			),
			'Esri Satellite': folium.TileLayer(
				tiles = 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
				attr = 'Esri',
				name = 'Esri Satellite',
				overlay = True,
				control = True
			)
		} 

	def draw(self, images, vis_params, layer_names = [], location=[20, 0], zoom_start = 3, height = 500):
		# draw using folium
		# Add EE drawing method to folium.
		folium.Map.add_ee_layer = self.add_ee_layer

		# Create a folium map object.
		self.m = folium.Map(location = location, zoom_start = zoom_start, height = height)

		# Add custom basemaps
		self.basemaps['Google Maps'].add_to(self.m)
		self.basemaps['Google Satellite Hybrid'].add_to(self.m)
		self.basemaps['Google Terrain'].add_to(self.m)

		# show coors interactively
		self.mouse_position()

		# # click show coors:
		# self._click_popup_coors()

		# click to place a popup marker
		self.m.add_child(folium.ClickForMarker(popup="Hello :)"))

		for image, vis_param, layer_name in zip(images, vis_params, layer_names):
			# Add the elevation model to the map object.
			self.m.add_ee_layer(image, vis_param, layer_name)

		# Add a layer control panel to the map.
		self.m.add_child(folium.LayerControl())

		# Add fullscreen button
		folium.plugins.Fullscreen().add_to(self.m)

		# # Display the map.
		# display(m)

	def save(self, filename):
		# example: m.save("filename.png")
		self.m.save(filename)

	# Define a method for displaying Earth Engine image tiles on a folium map.
	def add_ee_layer(self, ee_object, vis_params, name):
		
		try:    
			# display ee.Image()
			if isinstance(ee_object, ee.image.Image):    
				map_id_dict = ee.Image(ee_object).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
				).add_to(self.m)
			# display ee.ImageCollection()
			elif isinstance(ee_object, ee.imagecollection.ImageCollection):    
				ee_object_new = ee_object.mosaic()
				map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
				).add_to(self.m)
			# display ee.Geometry()
			elif isinstance(ee_object, ee.geometry.Geometry):    
				folium.GeoJson(
				data = ee_object.getInfo(),
				name = name,
				overlay = True,
				control = True
			).add_to(self.m)
			# display ee.FeatureCollection()
			elif isinstance(ee_object, ee.featurecollection.FeatureCollection):  
				ee_object_new = ee.Image().paint(ee_object, 0, 2)
				map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
			).add_to(self.m)
		
		except:
			print("Could not display {}".format(name))

	def mouse_position(self):
		formatter = "function(num) {return L.Util.formatNum(num, 3) + ' deg';};"

		folium.plugins.MousePosition(
			position='topright',
			separator=' | ',
			empty_string='NaN',
			lng_first=True,
			num_digits=20,
			prefix='Coordinates:',
			lat_formatter=formatter,
			lng_formatter=formatter,
		).add_to(self.m)

	def _click_popup_coors(self):
		folium.LatLngPopup().add_to(self.m)

	# easy visualizaiton parameter generator
	def viz(self, vmin, vmax, palette):
		return {
			"min": vmin,
			"max": vmax,
			"palette": palette
		}


import yaml
import zipfile
import multiprocessing
import sys
import itertools
import numbers
from pathlib import Path
from datetime import datetime, timedelta, date # date type is not datetime, it only accepts year, month and day.
from time import gmtime, strftime, ctime

class Montre(object):
	def __init__(self):
		super()

	def to_date(self, date_str, format = r"%Y-%m-%d"):
		return datetime.strptime(date_str, format)

	def to_str(self, cur_date, format = r"%Y-%m-%d"):
		return cur_date.strftime(format)

	# Check if the int given year is a leap year
	# return true if leap year or false otherwise
	def is_leap_year(self, year):
		if(year % 4) == 0:
			if(year % 100) == 0:
				if(year % 400) == 0:
					return True
				else:
					return False
			else:
				return True
		else:
			return False

	def manage_time(self, cur_date, years = 0, months = 0, weeks = 0, days = 0, hours = 0, minutes = 0, seconds = 0):
		# the finest scale is second.
		# input time must be datetime type
		if not isinstance(cur_date, datetime):
			raise(ValueError)
		# set output format
		format = r"%Y-%m-%d %H:%M:%S.%f"
		# disintegrate input time into subitems
		cur_year = cur_date.year
		cur_month = cur_date.month
		cur_day = cur_date.day
		cur_hour = cur_date.hour
		cur_minute = cur_date.minute
		cur_second = cur_date.second
		cur_ms = cur_date.microsecond
		# manage year add/substract
		if years != 0:
			cur_year = cur_year + years
		# mange month add/substract
		cur_month = cur_month + months
		if cur_month > 12:
			cur_year = int(cur_year + cur_month // 12)
			cur_month = int(cur_month % 12)

		if (cur_month == 2) and (cur_day > 28):
			if self.is_leap_year(cur_year):
				cur_day = 29
			else:
				cur_day = 28
		str_date = f"{cur_year}-{cur_month}-{cur_day} {cur_hour}:{cur_minute}:{cur_second}.{cur_ms}" 
		cur_date = datetime.strptime(str_date, format)
		# manage the rest
		delta_seconds = 0
		if weeks != 0:
			delta_seconds = delta_seconds +  weeks * 7 * 24 * 60 * 60
		if days != 0:
			delta_seconds = delta_seconds + days * 24 * 60 * 60
		if hours != 0:
			delta_seconds = delta_seconds + hours * 60 * 60
		if minutes != 0:
			delta_seconds = delta_seconds + minutes * 60
		if seconds != 0:
			delta_seconds = delta_seconds + seconds
		cur_date = cur_date + timedelta(seconds = delta_seconds)
		return cur_date

class Yaml(object):
    def __init__(self, path):
        super()
        self.path = path
        if isinstance(path, Path):
            self.path = self.path.as_posix() 

    def load(self):
        with open(self.path, "r") as stream:
            try:
                yamlfile = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                assert(exc)
        return yamlfile
    
    def dump(self, data_dict):
        with open(self.path, 'w') as f:
            yaml.dump(data_dict, f, default_flow_style = False)

# extract a .zip (p) into a folder at tar_dir of a name p.stem
def unzip(p, tar_dir, new_folder = True, folder_name = None, delete = False):
    # print(p)
    if not isinstance(p, Path):
        p = Path(p)
    if not isinstance(tar_dir, Path):
        tar_dir = Path(tar_dir)
    if new_folder:
        if not folder_name:
            out_dir = tar_dir.joinpath(p.stem)
        else:
            out_dir = tar_dir.joinpath(folder_name)
    else:
        out_dir = tar_dir
    # if not out_dir.exists(): out_dir.mkdir()
    create_all_parents(out_dir)
    with zipfile.ZipFile(p, "r") as zip_ref:
        zip_ref.extractall(out_dir)
    if delete:
        p.unlink()

# if parent or grand parent dirs not exist, 
# make them including current dir (if its not a file)
def create_all_parents(directory, flag = "a"):
    # flag: if directory is a dir ("d") or file ("f") or automatically desice ("a")
    if not isinstance(directory, Path):
        directory = Path(directory)
    parents = list(directory.parents)
    parents.reverse()
    # NOTICE: sometimes is_dir returns false, e.g., a dir of onedrive
    if flag == "a":
        if not directory.is_file():
            parents.append(directory)
    elif flag == "d":
        parents.append(directory)
    else:
        pass
    for p in parents:
      if not p.exists():
        p.mkdir()

# search current dir and list the subdirs
def searching_all_files(directory):
    if not isinstance(directory, Path):
        dirpath = Path(directory)
    assert(dirpath.is_dir())
    file_list = []
    for x in dirpath.iterdir():
        if x.is_file():
            file_list.append(x)
        elif x.is_dir():
            file_list.extend(searching_all_files(x))
    return file_list

# print progress bar at one line.
def pbar(idx, total, auto_check = False):
    if auto_check:
        if not isinstance(idx, numbers.Number):
            idx = float(idx)
        if not isinstance(total, numbers.Number):
            total = float(total)
    if (100*(idx + 1)/total).is_integer():
        sys.stdout.write(f"progress reaches {idx + 1} of {total}, {100*(idx + 1)/total}% ...\r")


# mount the drive
def mount_drive():
    from google.colab import drive
    from pathlib import Path
  
    drive.mount('/content/drive')
    return Path.cwd().joinpath('drive/My Drive')


# coding: utf-8

# author: sz394@exeter.ac.uk
# date: 22/10/2019

# description: this code sorts tob1 filenames and stores sorted into a text file.
# usage: Make_EdiRe_dirlist.py [-f] [--t]
#     -f: directory of folder containing input tob1s, the default is inputs;
#     --t: output text file directory, the default is output.txt;
# example: python Make_EdiRe_dirlist.py G:\workspace\Converted_Roth\Roth_1_N --t=RothN.txt --c=2000000
# python Make_EdiRe_dirlist.py G:\workspace\Roth\Converted_north\2017_11_02 --t=2017_11_02.txt --action=remove
# python Make_EdiRe_dirlist.py G:\workspace\Roth\Converted_north\2017_11_02 --t=2017_11_02.txt --action=daynight

import os, re, multiprocessing, argparse, shutil
from pathlib import Path

class EdiReObj(object):
    def __init__(self, tob1_folder_path):
        super()
        self.folder = tob1_folder_path

    def sort(self, t_dir, c, pool):
        print("ready to go...")
        # iterate the folder to find tob1 files.
        paths = Path(self.folder).glob(r"*.dat")
        # paths = list(paths) # if iterator doesnot work, uncomment this line.
        print("tob1 files iterated successfully...")

        # have a list of lists derived from the function split_filename
        list_of_splits = pool.map(split_filename, paths)
        # remove the filenames of which the size is smaller than the criterion.
        list_of_splits = list(filter(lambda x: x[6] >= c, list_of_splits))
        print("tob1 files loaded successfully...")
        # sort the filenames according to timestamps and only extract sorted filenames.
        list_of_splits.sort(key=lambda x: (x[1], x[2], x[3], x[4], x[5]))
        print("sorted successfully...")
        list_of_files = list(zip(*list_of_splits))[0]

        # write into files.
        with open(t_dir, 'w') as f:
            f.writelines("%s\n" % placeholder for placeholder in list_of_files)
        print("writting successfully, bye!")

    def remv_incoms(self, flag):
        '''
        remove the incomplete files from tob1 folder, incomplete files are those
        smaller than the max large or specified number.
        '''
        # iterate the folder to find tob1 files.
        paths = Path(self.folder).glob(r"*.dat")
        paths = list(paths) # generator cannot be used more than once.

        if flag == "max":
            print("Removing all incomplete files...")
            # fs_list = [p.stat().st_size for p in paths]
            fs_list = map(lambda p: p.stat().st_size, paths)
            max_size = max(fs_list)
        else:
            print("Removing user-defined size files specified by param c...")
            max_size = flag

        # codes here can be optimized!
        print(f"the max size is {max_size} bits.")
        for p in paths:
            if p.stat().st_size < max_size:
                print(f"{p.name} is being removed...")
                p.unlink() # remove file
                # p.rmdir() # remove the whole directory

    def daynight(self, sunrise, sunset, move):
        '''
        split files into diurnal and nocturnal, put them into different folders, respectively. 
        '''
        # iterate the folder to find tob1 files.
        paths = Path(self.folder).glob(r"*.dat")
        # create diurnal folder
        day_folder = Path.cwd().joinpath("day")
        day_folder.mkdir(parents=True, exist_ok=True)
        # create nocturnal folder
        night_folder = Path.cwd().joinpath("night")
        night_folder.mkdir(parents=True, exist_ok=True)

        # codes here can be optimized!
        print("start moving/copying day&night files....")
        for p in paths:
            pattern = r"\d{4}_\d{2}_\d{2}_\d{4}"
            results = re.findall(pattern, p.stem)
            if len(results) != 1:
                raise Exception("Wrong filename format!")
            _, _, _, HHMM = results[0].split("_")
            HH = int(HHMM[0: 2])
            if HH > sunrise and HH < sunset:
                # day
                to_file = day_folder.joinpath(p.name)
            else:
                to_file = night_folder.joinpath(p.name)
            if move == False:
                shutil.copy(p.as_posix(), to_file.as_posix())
            else:
                os.rename(p.as_posix(), to_file.as_posix())
        print("mission accomplised!")


    def __del__(self):
        pass

def split_filename(p):
    '''
    Use regex to find the timestamp information in one tob1 file,
    and split into YYYY, mm, dd, HH, MM, respectively.
    Return a list including filename and splitted time info.
    '''
    filesize = p.stat().st_size
    filename = p.as_posix()
    pattern = r"\d{4}_\d{2}_\d{2}_\d{4}"
    results = re.findall(pattern, filename)
    if len(results) != 1:
        raise Exception("Wrong filename format!")
    YYYY, mm, dd, HHMM = results[0].split("_")
    YYYY = int(YYYY)
    mm = int(mm)
    dd = int(dd)
    HH = int(HHMM[0: 2])
    MM = int(HHMM[2::])
    return [filename, YYYY, mm, dd, HH, MM, filesize]

def main() :
    # define command line params.
    parser = argparse.ArgumentParser(description = "please specify directories of folder containing to be sorted tob1 files and of output txt file.")
    parser.add_argument('f', type=str, default = "inputs",
                        help="directory of folder containing input tob1s.")

    parser.add_argument('--t', type=str, default = "output.txt",
                        help="output text file directory.")

    parser.add_argument('--c', type=int, default = 10, #bits
                        help="criterion")

    parser.add_argument('--action', type=str, default = "sort",
                        help="action to do with tob1s")

    parser.add_argument('--sunrise', type=int, default = 6,
                        help="time of sunrise")

    parser.add_argument('--sunset', type=int, default = 18,
                        help="time of sunset, 24-hour system")

    parser.add_argument('--move', type=bool, default = False,
                        help="if move or copy day/night tob1 files")


    args = parser.parse_args()
    
    # transform Windows directories into posix
    if "\\" in args.f:
        folder = args.f.replace("\\", "/")
    else:
        folder = args.f

    if "\\" in args.t:
        t_dir = args.t.replace("\\", "/")
    else:
        t_dir = args.t

    # check the capability of multiprocessing and create the pool.
    cores = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(processes = cores)

    # initialize object
    eddy = EdiReObj(folder)
    if args.action == "sort":
        eddy.sort(t_dir, args.c, pool)

    elif args.action == "remove":
        if args.c == parser.get_default('c'):
            eddy.remv_incoms("max")
        else:
            eddy.remv_incoms(args.c)
    elif args.action == "daynight":
        eddy.daynight(args.sunrise, args.sunset, args.move)
        
    else:
        raise Exception("Not identified action!")



if __name__ == "__main__":
    main()


# coding: utf-8
__all__ = ["scigeo", "fluxlib"]

# from pysy.scigeo import *
# from pysy.scigee import *
# from pysy.fluxlib import *
# from pysy.toolbox import *
import ee

ee.Initialize()

from . import scigee
__all__ += ["scigee"]

from .scigee import *
__all__ += scigee.__all__

from . import  toolbox
__all__ += toolbox.__all__

from .toolbox import *
__all__ += toolbox.__all__


def gen_buffer(lon, lat, step, shape = "rectangle"):
    if shape == "rectangle":
        # clockwise
        coors = [
                 [lon - step, lat + step], # upper left
                 [lon + step, lat + step], # upper right
                 [lon + step, lat - step], # lower right
                 [lon - step, lat - step], # lower left
        ]

    return coors


import pickle
import multiprocessing
import numpy as np
import pandas as pd
from pathlib import Path
from pysy.toolbox.sysutil import Yaml, unzip, create_all_parents, pbar

class Fluxnet2015(object):
    def __init__(self, path_cfg):
        super()
        fyaml = Yaml(path_cfg)
        self.config = fyaml.load()
        self.root = self.config["dirs"]["root"] # root dir for all fluxnet2015 data
        if not isinstance(self.root, Path):
            self.root = Path(self.root)

    def untar(self):
        # initialize multicores
        cores = multiprocessing.cpu_count()
        pool  = multiprocessing.Pool(processes=cores)

        # unzip use multiprocessors 
        src_dir = self.root.joinpath(self.config["dirs"]["src"])
        dst_dir = self.root.joinpath(self.config["dirs"]["dst"])
        create_all_parents(dst_dir)
        paths = list(Path(src_dir).glob(r"*.zip"))
        lengs = len(paths)
        out_dirs = [dst_dir for leng in range(lengs)]
        print(f"core number is: {cores}")
        pool.starmap(unzip, zip(paths, out_dirs))
    
    def get_meta(self):
        # generate necessary info, extract site ID, longitude, latitude as a dataframe
        meta_dir = self.root.joinpath(self.config["dirs"]["meta"])
        meta = pd.read_excel(meta_dir, usecols = ["SITE_ID", "VARIABLE", "DATAVALUE"])
        lats = meta[meta["VARIABLE"] == "LOCATION_LAT"]
        lats = lats.rename(columns={"DATAVALUE": "LAT"})
        lons = meta[meta["VARIABLE"] == "LOCATION_LONG"]
        lons = lons.rename(columns={"DATAVALUE": "LON"})
        meta = pd.merge(lats, lons, on='SITE_ID')
        self.meta = meta.loc[:, ["SITE_ID", "LAT", "LON"]]
        # print(meta)
        # site_ids = meta["SITE_ID"].values
        # print(site_ids)
        print(f"meta data was accessed...")

    def search_file(self):
        freq = self.config["params"]["freq"]
        dst_dir = self.root.joinpath(self.config["dirs"]["dst"])
        paths = Path(dst_dir).rglob("*.csv")
        # print(list(paths))

        # paths = [p for p in paths if "HH" in p.stem.split("_")]
        paths_ = []
        for p in paths:
            temp = p.stem.split("_")
            if freq == "HH":
                if "HH" in temp or "HR" in temp: # HR rather than HH in US-UMB
                    paths_.append(p)
            if freq == "DD":
                if "DD" in temp:
                    paths_.append(p)
            if freq == "MM":
                if "MM" in temp:
                    paths_.append(p)
        paths = paths_
        site_names_in_paths = np.array([Path(p).stem.split("_")[1] for p in paths])

        data_dict = {}

        # iterate each row (each site) of meta which is a dataframe
        for index, row in self.meta.iterrows():
            print(index, row["SITE_ID"])
            si = row["SITE_ID"]
            data_dict[si] = {}
            data_dict[si]["LON"] = row["LON"]
            data_dict[si]["LAT"] = row["LAT"]
            idxs = np.where(site_names_in_paths == si)[0]
            data_dict[si]["PATHS"] = [paths[i].as_posix() for i in idxs]
            # print(data_dict)
            # exit(0)

        site_info_dir = self.root.joinpath(self.config["dirs"]["site_info"])
        fyaml = Yaml(site_info_dir)
        fyaml.dump(data_dict)
        print(f"fluxnet files were iterated...")
        
    def retrieve(self):
        site_names = self.config["params"]["site_name"]
        savefile = self.root.joinpath(self.config["dirs"]["fpkl"])

        site_info_dir = self.root.joinpath(self.config["dirs"]["site_info"])
        fyaml = Yaml(site_info_dir)
        info = fyaml.load()

        if not site_names:
            site_names = list(info.keys())

        vars_req = self.config["params"]["vars"]
        # read each fluxnet file
        records = {}
        for count, (site_name, cell) in enumerate(info.items()):
            if not site_name in site_names:
                continue
            print(site_name)
            lat = cell["LAT"]
            lon = cell["LON"]
            paths = cell["PATHS"]
            # aux = pd.read_csv(paths[0]) # ERAI auxiliary data
            data = pd.read_csv(paths[1]) # fullset data
            # check the variable names
            # print(aux.columns.values)
            # print(data.columns.values)
            try:
                # print the time range of each fluxnet data
                print(Path(paths[1]).stem.split("_")[-2])
            except Exception as e:
                print(e)
                print(paths)
            df = data[vars_req]
            pbar(count, len(info.items()))
            item = {
                "lat":  lat,
                "lon": lon,
                "values": df
            }
            records[site_name] = item
        with open(savefile, "wb") as f:
            pickle.dump(records, f, protocol=pickle.HIGHEST_PROTOCOL)



import os
import ee
import requests
import numpy as np

class EEarth(object):
    """
    Earth engine object
    """
    def __init__(self, dataset_name):
        super()
        ee.Initialize()
        self.dataset_name = dataset_name

    def fetch_image(self):
        self.image = ee.Image(self.dataset_name)

    def fetch_collection(self, date_range = [], roi = None):
        collection = ee.ImageCollection(self.dataset_name)
        # filter by date:
        if date_range:
            start_date, end_date = date_range
            collection = collection.filterDate(start_date, end_date)
        # filter by bounds:
        if roi:
            collection = collection.filterBounds(roi)
        self.__length__ = collection.size().getInfo()
        self.collection = collection

    # def fetch_collection(self, date_range = [], roi = None, to_list = False):
    # 	collection = ee.ImageCollection(self.dataset_name)
    # 	# filter by date:
    # 	if date_range:
    # 		start_date, end_date = date_range
    # 		collection = collection.filterDate(start_date, end_date)
    # 	# filter by bounds:
    # 	if roi:
    # 		collection = collection.filterBounds(roi)
    # 	self.__length__ = collection.size().getInfo()
    # 	# covert ee collection to list?
    # 	if to_list: 
    # 		collection = collection.toList(collection.size())
    # 	self.collection = collection
    
    # def get_image_by_index(self, idx):
    # 	self.image = ee.Image(self.collection.get(idx))

    # def reduce_collection(self, band = None, label = "DATA", band_reducer = ee.Reducer.mean(), spatial_reducer = ee.Reducer.median()):
    # 	if band:
    # 		image = self.collection.select(band).reduce(spatial_reducer).rename(band)
    # 	else:
    # 		image = self.collection.map(
    # 				lambda image: image.reduce(band_reducer)
    # 			).reduce(spatial_reducer).rename(label)
    # 	self.image = image

class Ecolbox(EEarth):
    """
    Ecolbox: Earth engine Collection Toolbox
    """
    def __init__(self, collection_name, date_range = [], roi = None):
        EEarth.__init__(self, collection_name)
        self.fetch_collection(date_range, roi)
        self.__length__ = self.collection.size().getInfo()

    def __call__(self, collection):
        # update collection box
        self.__length__ = collection.size().getInfo()
        self.collection = collection

    def __to_ee_list(self):
        self.__image_list__ = self.collection.toList(self.collection.size())

    def to_list(self, collection):
        if not collection:
            collection = self.collection
        return collection.toList(collection.size()).getInfo()

    def get_image_by_index(self, collection = None, idx = 0):
        # default: first.
        if collection:
            collection = collection.toList(collection.size())
            image = ee.Image(collection.get(idx))
        else:
            collection = self.collection
            if not isinstance(collection, ee.ee_list.List):
                self.__to_ee_list()
                image = ee.Image(self.__image_list__.get(idx))
        return image

    # def fmap(self, func, *args,**kwargs):
    #     # simplify the map function in the future
    #     return self.collection.map(
    #         lambda image: func(image, *args,**kwargs)
    #     )

    def reduce_collection(self, collection = None, band = None, label = "DATA", band_reducer = ee.Reducer.mean(), spatial_reducer = ee.Reducer.median()):
        if not collection:
            collection = self.collection
        if band:
            image = collection.select(band).reduce(spatial_reducer).rename(band)
        else:
            image = collection.map(
                    lambda image: image.reduce(band_reducer)
                ).reduce(spatial_reducer).rename(label)
        return image

class Emagebox(object):
    """
    Emagebox: Earth engine Image Toolbox
    """
    def __init__(self, image, scale = 30, max_pixels = 1e8, default_value = -9999):
        self.image = image
        self.scale = scale
        self.max_pixels = max_pixels
        self.default_value = default_value

    def set_scale(self, scale):
        self.scale = scale
    
    def set_max_pixels(self, max_pixels):
        self.max_pixels = max_pixels
    
    def set_default_value(self, default_value):
        self.default_value = default_value

    def get_band_names(self):
        return self.image.bandNames().getInfo()

    def get_stats(self, roi, reducer = ee.Reducer.mean()):
        stat = self.image.reduceRegion(
            reducer = reducer,
            geometry = roi,
            scale = self.scale,
            maxPixels = self.max_pixels
        )
        return stat.getInfo()
    
    def get_date(self):
        date = ee.Date(self.image.get('system:time_start'))
        return date.format('Y-M-d').getInfo()
    
    def get_proj(self, band, get_info = False):
        # // Get projection information from band.
        proj = self.image.select(band).projection()
        if get_info:
            proj = proj.getInfo()
        return proj

    def get_scale(self, band):
        # // Get scale (in meters) information from band.
        scale = self.image.select(band).projection().nominalScale()
        return scale.getInfo()

    def reproject(self, proj):
        return self.image.reproject(proj)

    def clip(self, roi):
        return self.image.clip(roi)

    def mask_value(self):
        mask = self.image.eq(self.default_value)
        self.image = self.image.updateMask(mask)

    def unmask(self):
        self.image = self.image.unmask(ee.Image.constant(self.default_value))

    def get_value(self, band, point, reducer = ee.Reducer.first()):
        value = self.image.select(band).reduceRegion(reducer, point, self.scale).get(band)
        value = ee.Number(value)
        return value.getInfo()

    def get_values(self, band, roi):
        # getInfo() is limited to 5000 records
        # ee.ee_exception.EEException: Array: No numbers in 'values', must provide a type.
        # see more at: https://gis.stackexchange.com/questions/321560/getting-dem-values-as-numpy-array-in-earth-engine
        # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
        # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
        latlng = ee.Image.pixelLonLat().addBands(self.image.clip(roi).unmask(ee.Image.constant(self.default_value)))
        latlng = latlng.reduceRegion(
            reducer = ee.Reducer.toList(), 
            geometry = roi, 
            maxPixels = self.max_pixels, 
            scale = self.scale
        )
        lats = np.array((ee.Array(latlng.get("latitude")).getInfo()))
        lons = np.array((ee.Array(latlng.get("longitude")).getInfo()))
        try:
            values = np.array((ee.Array(latlng.get(band)).getInfo()))
        except:
            values = np.full_like(lats, np.nan, dtype = np.float64)
        # self.values = list(values) ## print as list to check

        if not (values.shape == lats.shape == lons.shape):
            raise Exception(
                f"SizeError: " +
                f"values shape is {values.shape}, " + 
                f"lats shape is {lats.shape}, " + 
                f"lons shape is {lons.shape}."
            )
        return {
            "values":  values,
            "lons": lons,
            "lats": lats
        }

    def localize(self, save_name, save_folder = ".", crs_epsg = "4326"):
        url = self.image.getDownloadURL({
            "name": save_name,
            "crs": "EPSG:" + crs_epsg,
            "scale": self.scale
        })
        if not os.path.exists(save_folder): os.makedirs(save_folder)
        save_dir = f"{save_folder}/{save_name}.zip"

        # Download the subset
        r = requests.get(url, stream = True)
        with open(save_dir, 'wb') as fd:
            for chunk in r.iter_content(chunk_size = 1024):
                fd.write(chunk)

class VI(object):
    def __init__(self):
        pass

    def __call__(self, vi, image, **kwargs):
        if vi == "evi":
            image = self.calc_evi(image, **kwargs)
        elif vi == "cire":
            image = self.calc_cire(image, **kwargs)

        return image

    @staticmethod
    def calc_evi(image, **kwargs):
        nir = kwargs["nir"]
        red = kwargs["red"]
        blue = kwargs["blue"]

        evi = image.expression(
            '2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))', {
            'NIR': image.select(nir),
            'RED': image.select(red),
            'BLUE': image.select(blue)
        })
        return image.addBands(evi.rename("EVI"))

    @staticmethod
    def calc_ndi(image, **kwargs):
        band_a = kwargs["b1"]
        band_b = kwargs["b2"]

        ndi = image.expression(
            '(band_a - band_b) / (band_a + band_b)', {
            'band_a': image.select(band_a),
            'band_b': image.select(band_b)
        })
        if "label" in kwargs.keys():
            label = kwargs["label"]
        else:
            label = "NDI"
        return image.addBands(ndi.rename(label))

    @staticmethod
    def calc_cire(image, **kwargs):
        re2 = kwargs["re2"]
        re3 = kwargs["re3"]

        cire = image.expression(
            '(RE3/RE2) - 1', {
            'RE2': image.select(re2),
            'RE3': image.select(re3)
        })
        return image.addBands(cire.rename("CIRE"))

class Point(object):
    def __init__(self):
        pass

    def __call__(self, bounds = None, lon = None, lat = None):
        if bounds:
            lon, lat = bounds
        assert lon, "No longitude input..."
        assert lat, "No latitude input..."
        point = ee.Geometry.Point(lon, lat)
        return point

    @classmethod
    def get_circle_buffer(self, point, buffer_size = 100):
        return point.buffer(buffer_size)

    @classmethod
    def get_rect_buffer(self, point, buffer_size = 0.5):
        # buffer_size unit eq proj unit, e.g., degree for WGS84
        lon, lat = point.getInfo()["coordinates"]
        # example: [-97.94, 26.81, -96.52, 26.84]
        bounds = [lon - buffer_size, lat - buffer_size, lon + buffer_size, lat + buffer_size]
        return ee.Geometry.Rectangle(bounds)

class Polygon(object):
    def __init__(self):
        pass

    def __call__(self, bounds):
        # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
        # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]

        if isinstance(bounds, list):
            if np.array(bounds).ndim == 1:
                region = ee.Geometry.Rectangle(bounds)
            else:
                region = ee.Geometry.Polygon(bounds)
        return region

class Geometry(Point, Polygon):
    def __init__(self, bounds = None, lon = None, lat = None):
        Point.__init__(self)
        Polygon.__init__(self)
        self.bounds = bounds
        self.lon = lon
        self.lat = lat

    def __call__(self, geom_type):
        if geom_type == 0 or geom_type == "point":
            geom = Point.__call__(self, bounds = self.bounds, lon = self.lon, lat = self.lat)
        elif geom_type == 2 or geom_type == "polygon":
            geom = Polygon.__call__(self, bounds = self.bounds)
        return geom


class Utils(VI):
    def __init__(self):
        VI.__init__(self)

    @classmethod
    def calc_vi(self, vi, image, **kwargs):
        return VI.__call__(self, vi, image, **kwargs)

    @classmethod
    def sentinel_2_cloud_mask(self, image, mask_out = True):
        """
        javascript code:

        # function maskS2clouds(image) {
        #   var qa = image.select('QA60');

        #   // Bits 10 and 11 are clouds and cirrus, respectively.
        #   var cloudBitMask = 1 << 10;
        #   var cirrusBitMask = 1 << 11;

        #   // Both flags should be set to zero, indicating clear conditions.
        #   var mask = qa.bitwiseAnd(cloudBitMask).eq(0)
        #       .and(qa.bitwiseAnd(cirrusBitMask).eq(0));

        #   return image.updateMask(mask).divide(10000);
        # }

        European Space Agency (ESA) clouds from 'QA60', i.e. Quality Assessment band at 60m
        
        parsed by Nick Clinton
        """

        qa = image.select('QA60')

        # bits 10 and 11 are clouds and cirrus
        cloudBitMask = int(2**10)
        cirrusBitMask = int(2**11)

        # both flags set to zero indicates clear conditions.
        mask = qa.bitwiseAnd(cloudBitMask).eq(0).And(\
            qa.bitwiseAnd(cirrusBitMask).eq(0))
        
        if mask_out:
            return image.updateMask(mask)#.divide(10000)
        else:
            # clouds is not clear
            cloud = mask.Not().rename(['ESA_clouds'])

            # return the masked and scaled data.
            return image.addBands(cloud)

    # @classmethod
    # def coors2roi(self, bounds):
    #     # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
    #     # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
    #     if isinstance(bounds, list):
    #         if np.array(bounds).ndim == 1:
    #             roi = ee.Geometry.Rectangle(bounds)
    #         else:
    #             roi = ee.Geometry.Polygon(bounds)
    #     return roi

    # @classmethod
    # def coor2point(self, coors = None, lon = None, lat = None):
    #     if coors:
    #         lon, lat = coors
    #     assert lon, "No longitude input..."
    #     assert lat, "No latitude input..."
    #     point = ee.Geometry.Point(lon, lat)
    #     return point

    # @classmethod
    # def get_circle_buffer(self, point, buffer_size = 100):
    #     return point.buffer(buffer_size)

    # @classmethod	
    # def get_rect_buffer(self, point, buffer_size = 0.5):
    #     # buffer_size unit eq proj unit, e.g., degree for WGS84
    #     lon, lat = point.getInfo()["coordinates"]
    #     # example: [-97.94, 26.81, -96.52, 26.84]
    #     bounds = [lon - buffer_size, lat - buffer_size, lon + buffer_size, lat + buffer_size]
    #     return ee.Geometry.Rectangle(bounds)


import yaml, argparse
from pathlib import Path
from utilizes import EddyProc


def main():
    parser = argparse.ArgumentParser(description = "Do you have a complete yaml proclist or prefer param configuration?")
    parser.add_argument('d', type=str, help="yaml dir")
    parser.add_argument('--t', type=str, default = "p", help="transcript type, p for param configured, c for complete.")
    parser.add_argument('--o', type=str, default="./output", help="output folder")

    args = parser.parse_args()
    # yaml_dir = "half_complete_proc_list.yaml"
    # yaml_dir = "config.yaml"
    yaml_dir = args.d
    out_folder = Path(args.o)
    # print(out_folder)
    if not out_folder.is_absolute():
        if ".." in [p.as_posix() for p in out_folder.parents]:
            out_folder = Path.cwd().parent.joinpath(out_folder.name)
        else:
            out_folder = Path.cwd().joinpath(out_folder.name)
            # debug:
            # print("no ..")
            # print(list(out_folder.parents))
            pass
    else:
        # debug:
        # print("it's abs")
        pass

    parents = list(out_folder.parents)
    parents.append(out_folder)
    parents.reverse()
    for par in parents:
        if not par.exists(): 
            par.mkdir()
        else:
            print(f"{par} exists")

    eddy = EddyProc()
    print("Starting...")
    with open(yaml_dir, 'r') as stream:
        try:
            user_config = yaml.safe_load(stream)
            if args.t == 'c':
                transcrpit_yaml_list(user_config, eddy, out_folder)
            elif args.t == 'p':
                parse_yaml_params(user_config, eddy, out_folder)
            else:
                raise Exception("please specify the correct transcript type!")
        except yaml.YAMLError as exc:
            print(exc)
    
    print("Finished successfully.")

def parse_yaml_params(inyaml, eddy, out_folder, outfile = "proc_list.txt", outyaml = "variables.yaml"):

    invars = inyaml["vars"]
    if "pos" in invars.keys():
        # do something...
        channels = invars["pos"]
        pass
    elif "pos_begin" in invars.keys():
        channels = [invars["pos_begin"] + idx for idx in range(len(invars["labels"]))]
    else:
        raise Exception("No channel info!")
    # dCalc = inyaml["outdir"]["dCalc"]
    # dSpec = inyaml["outdir"]["dSpec"]
    results = "" # change to "".join(list)
    cur_vars = {}
    cur_vars["vars_chns"] = dict(zip(invars["labels"], channels))
    cur_vars["vars_chns"].update(inyaml["outdir"])
    cur_vars["in_vars"] = invars["labels"]
    ######## START PROC LIST
    results += eddy.location_output_files(dCalc = inyaml["outdir"]["dCalc"], dSpec= inyaml["outdir"]["dSpec"])
    results += eddy.comments(cmt1 = "Extract the data")
    for idx, pos in enumerate(channels):
        results += eddy.extract(chn = pos, label = invars["labels"][idx])
    results += eddy.comments(cmt1 = "Convert cell (c) and block (b) RH's to absolute densities")
    for label in ["cH2O", "c_e"]:
        results += eddy.extract(chn = cur_vars["vars_chns"]["cRH"], label = label)
        cur_vars["vars_chns"][label] = cur_vars["vars_chns"]["cRH"]
        cur_vars["in_vars"].append(label)
    for label in ["bH2O", "b_e"]:
        results += eddy.extract(chn = cur_vars["vars_chns"]["bRH"], label = label)
        cur_vars["vars_chns"][label] = cur_vars["vars_chns"]["bRH"]
        cur_vars["in_vars"].append(label)
    params = inyaml["humidity_unit_conv"] # intro callback function instead!!!.
    for param in params:
        results += eddy.gas_conversion_time_series(**param)
    results += eddy.comments(cmt1 = "Clean data")
    results += eddy.despike(sig = "Ts", replaceSpikes = "x")
    cur_vars["new_vars"]= ["x"]
    results += eddy.despike_vickers(sig = "Ts", count_ts_vickers = "count_Ts_Vickers")
    cur_vars["new_vars"].append("count_Ts_Vickers")
    results += eddy.comments(cmt1 = "Calculate some means")
    for var in cur_vars["in_vars"]:
        if var in ["Ux", "Uy", "Uz"]:
            new_mean_var = "mean_" + var + "_pre_rot"
            results += eddy.one_chn_statistics(sig = var, SLmean = new_mean_var)
            cur_vars["new_vars"].append(new_mean_var)
        elif var == "Ts":
            new_mean_var = "mean_" + var + "_virt"
            results += eddy.one_chn_statistics(sig = var, SLmean = new_mean_var)
            cur_vars["new_vars"].append(new_mean_var)
        elif var not in ["c_e", "b_e"]:
            new_mean_var = "mean_" + var
            results += eddy.one_chn_statistics(sig = var, SLmean = new_mean_var)
            cur_vars["new_vars"].append(new_mean_var)
    results += eddy.virtual_temperature_raw(sig = "Ts", sigH2O = "bH2O", PKpa = "mean_cP")
    for var in ["Ts", "c_e", "b_e"]:
            new_mean_var = "mean_" + var
            results += eddy.one_chn_statistics(sig = var, SLmean = new_mean_var)
            cur_vars["new_vars"].append(new_mean_var)
    results += eddy.comments(cmt1 = "Plot some of the means")
    params = inyaml["plot_mean"]
    for param in params:
        results += eddy.plot_value(**param)
    results += eddy.comments(
        cmt1 = "Calculate the wind direction before rotations",
        cmt2 = "Wind speed is calculated as part of the coordinate rotation",
        cmt3 = "Save Ux, Uy, and Uz Components before rotation"
    )
    cur_vars["new_vars"].append("Wind_Dir")
    cur_vars["new_vars"].append("Wind_Dir_Std")
    results += eddy.wind_direction(sigU="Ux", sigV="Uy", orient="0", SLWindDir = "Wind_Dir", SLWindDirDiv = "Wind_Dir_Std")
    results += eddy.comments("X corrs")

    params = inyaml["cross_corr"]
    for param in params:
        cur_vars["new_vars"].append(param["SLPeakTime"])
        results += eddy.cross_correlate(**param)
        results += eddy.plot_cross_correlation()
    
    results += eddy.comments("Remove Lags")

    params = inyaml["remove_lag"]
    for param in params:
        results += eddy.remove_lag(**param)

    results += eddy.comments("Coordinate Rotation")

    results += eddy.rotation_coefficients(sigU="Ux", sigV="Uy", sigW="Uz")

    results += eddy.rotation(sigU="Ux", sigV="Uy", sigW="Uz", do1Rot="x", do2Rot="x")

    results += eddy.one_chn_statistics(sig="Ux", SLmean="mean_Wind_Spd")
    cur_vars["new_vars"].append("mean_Wind_Spd")

    results += eddy.comments("Plot Spectra and Cospectra")

    for sig in ["Ts", "bH2O", "cH2O", "vCO2"]:
        results += eddy.spectra(sig=sig)
        results += eddy.cospectra(sig1="Uz", sig2=sig)
        results += eddy.plot_spectral()

    results += eddy.comments("CO2 from ppm to umol per m3")

    params = inyaml["co2_umol2density"]
    for param in params:
        results += eddy.gas_conversion_time_series(**param)

    results += eddy.comments("Calculate fluxes")

    results += eddy.sensible_heat_flux_coefficient(
        SLable="rhocp", vaporPres="mean_b_e", temp="mean_Ts", pres="mean_cP"
        )
    cur_vars["new_vars"].append("rhocp")

    results += eddy.latent_heat_of_evaporation(
        SLable="lambda", temp = "mean_Ts", pres = "mean_cP"
        )
    cur_vars["new_vars"].append("lambda")

    results += eddy.friction_velocity(
        sigU="Ux", sigV="Uy", sigW="Uz", SLuwvw = "friction_velocity"
        )
    cur_vars["new_vars"].append("friction_velocity")

    results += eddy.plot_value(LVal = "mean_Wind_Spd", RVal = "friction_velocity")

    params = inyaml["cal_flux_coef"]
    for param in params:
        results += eddy.two_chn_statistics(**param)
        cur_vars["new_vars"].append(param["SLFlux"])

    params = inyaml["plot_flux"]
    for param in params:
        results += eddy.plot_value(**param)

    results += eddy.comments("Frequency response correction")

    params = inyaml["tube_atten"]
    for param in params:
        results += eddy.tube_attenuation(**param)
        cur_vars["new_vars"].append(param["SLable"])

    param = inyaml["set_values"]
    cur_vars["set_values"] = param
    results += eddy.set_value(valDict = param)
    # for p_item in param.items():
    #     cur_vars["set_values"][p_item[0]] = p_item[1]

    results += eddy.mathematical_operation(
        SLable = "irga_path", measuredVarA = "mean_Wind_Spd",
        operation = "*", measuredVarB = "0.1"
    )
    cur_vars["new_vars"].append("irga_path")

    results += eddy.user_defined(
        SLable = "irga_path2", 
        equation = "Max(irga_path,0.005)",
        var = "irga_path"
    )
    cur_vars["new_vars"].append("irga_path2")

    param = inyaml["monin_obhukov"]
    results += eddy.stability_monin_obhukov(**param)
    cur_vars["new_vars"].append(param["SLable"])
    results += eddy.plot_value(LVal = "MO_stability")

    params = inyaml["freq_resp"]
    for param in params:
        results += eddy.frequency_response(**param)
        cur_vars["new_vars"].append(param["SLable"])
    
    params = inyaml["plot_freq_resp"]
    for param in params:
        results += eddy.plot_value(**param)

    params = inyaml["math_freq_oper"]
    for param in params:
        results += eddy.mathematical_operation(**param)
        cur_vars["new_vars"].append(param["SLable"])

    params = inyaml["plot_math_freq_oper"]
    for param in params:
        results += eddy.plot_value(**param)

    # Write lines and variables out

    with open(out_folder.joinpath(outfile), "w") as f:
        f.write(results[0: -1])

    temp = {}
    temp["attr"] = {
        "vars": {
            "in_vars": cur_vars["in_vars"],
            "new_vars": cur_vars["new_vars"]
        },
        "user_define": {
            "vars_chns": cur_vars["vars_chns"],
            "set_values": cur_vars["set_values"]
        }
    }
    cur_vars = temp

    with open(out_folder.joinpath(outyaml), 'w') as f:
        yaml.dump(cur_vars, f, default_flow_style = False)

    # debug:
    # print(list(cur_vars.keys()))
    print("done")

def transcrpit_yaml_list(yamlfile, eddy, out_folder, outfile = "transcripted_list.txt"):
    results = ""
    blocks = yamlfile["eddy_proc_list"]
    # print(blocks)
    # vars_dict = {}
    # vars_dict.update()
    for block in blocks:
        func_name = block[0]
        func_vars = block[1]
        # print(func_vars)
        # print(func_name)
        # print(func_vars)
        func = getattr(eddy, func_name)
        for func_var in func_vars:
            # print(func_var)
            results += func(**func_var)
    #     exit(0)
    # print(results)
    with open(out_folder.joinpath(outfile), "w") as f:
        f.write(results[0: -1])
    print("done")

if __name__ == "__main__":
    main()